import numpy as np
import pandas as pd
from scipy.stats import norm
import multiprocessing
from scipy.stats import chi2
from sklearn.decomposition import PCA
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.model_selection import train_test_split
from sklearn.metrics import mean_squared_error
from kneed import KneeLocator
from ete3 import Tree
from sklearn.model_selection import train_test_split

# from xgboost import XGBClassifier
# import allel
from functools import partial
import sys
from collections import Counter

from snpio.plotting.plotting import Plotting
from snpio.filtering.nremover2 import NRemover2 as NRemover
from sklearn.impute import SimpleImputer
from sklearn.preprocessing import OneHotEncoder


class PopGenStatistics:
    def __init__(self, popgenio):
        raise NotImplemented("PopGenStatistics has not yet been implemented.")
        if popgenio.filtered_alignment is not None:
            self.alignment = popgenio.filtered_alignment
        else:
            self.alignment = popgenio.alignment
        self.popmap = popgenio.populations
        self.populations = popgenio.populations
        self.plotting = Plotting(popgenio)
        self.nremover = NRemover(popgenio)

    def patterson_d_statistic_permutation(self, d1, d2, d3, outgroup):
        alignment_array = np.array(
            [list(rec) for rec in self.alignment], dtype=str
        )

        d1_base = alignment_array[d1, :]
        d2_base = alignment_array[d2, :]
        d3_base = alignment_array[d3, :]
        outgroup_base = alignment_array[outgroup, :]

        not_gap_mask = ~np.logical_or.reduce(
            (
                d1_base == "-",
                d2_base == "-",
                d3_base == "-",
                outgroup_base == "-",
            )
        )

        abba_mask = np.logical_and(
            d1_base == d2_base, d3_base == outgroup_base
        )
        baba_mask = np.logical_and(
            d1_base == d3_base, d2_base == outgroup_base
        )

        abba_count = np.sum(np.logical_and(abba_mask, not_gap_mask))
        baba_count = np.sum(np.logical_and(baba_mask, not_gap_mask))

        total_count = abba_count + baba_count

        d_statistic = (abba_count - baba_count) / total_count
        return d_statistic

    def partitioned_d_statistic_permutation(self, d1, d2, d3, d4, outgroup):
        alignment_array = np.array(
            [list(rec) for rec in self.alignment], dtype=str
        )

        d1_base = alignment_array[d1, :]
        d2_base = alignment_array[d2, :]
        d3_base = alignment_array[d3, :]
        d4_base = alignment_array[d4, :]
        outgroup_base = alignment_array[outgroup, :]

        not_gap_mask = ~np.logical_or.reduce(
            (
                d1_base == "-",
                d2_base == "-",
                d3_base == "-",
                d4_base == "-",
                outgroup_base == "-",
            )
        )

        abcd_mask = np.logical_and(
            d1_base == d2_base,
            np.logical_and(d3_base == d4_base, d3_base != outgroup_base),
        )
        dcba_mask = np.logical_and(
            d1_base == d4_base,
            np.logical_and(d2_base == d3_base, d2_base != outgroup_base),
        )

        abcd_count = np.sum(np.logical_and(abcd_mask, not_gap_mask))
        dcba_count = np.sum(np.logical_and(dcba_mask, not_gap_mask))

        total_count = abcd_count + dcba_count

        partitioned_d_statistic = (abcd_count - dcba_count) / total_count
        return partitioned_d_statistic

    def calculate_z_scores_and_p_values(self, d_statistic_values):
        d_statistic_array = np.array(list(d_statistic_values.values()))

        mean = np.mean(d_statistic_array)
        std_dev = np.std(d_statistic_array, ddof=1)

        z_scores = {
            combination: (value - mean) / std_dev
            for combination, value in d_statistic_values.items()
        }
        p_values = {
            combination: norm.sf(abs(z_score))
            for combination, z_score in z_scores.items()
        }

        # Apply Bonferroni correction
        num_tests = len(p_values)
        adjusted_p_values = {
            combination: min(1, p_value * num_tests)
            for combination, p_value in p_values.items()
        }

        result = {
            combination: {
                "z_score": z_scores[combination],
                "p_value": adjusted_p_values[combination],
            }
            for combination in d_statistic_values.keys()
        }
        return result

    def calculate_site_patterns(self, p1, p2, p3, p4, outgroup):
        """
        Calculate the site patterns from the given samples.

        Args:
            p1, p2, p3, p4 (str): Sample IDs.
            outgroup (str): Outgroup sample ID.

        Returns:
            dict: Dictionary of site patterns counts.
        """
        site_patterns = {"ABBA": 0, "BABA": 0, "BBAA": 0, "AAAB": 0, "AABB": 0}
        index_map = {p1: 0, p2: 1, p3: 2, p4: 3, outgroup: 4}
        indices = list(index_map.values())

        alignment_array = np.array([list(site) for site in self.alignment])

        for site in alignment_array[:, indices]:
            alleles = site.tolist()

            # Skip sites with missing or ambiguous data
            if "-" in alleles or "N" in alleles:
                continue

            # Calculate site pattern
            pattern = "".join(
                ["A" if a == alleles[4] else "B" for a in alleles[:4]]
            )

            # Increment the corresponding site pattern count
            if pattern in site_patterns:
                site_patterns[pattern] += 1

        return site_patterns

    def bootstrap_alignment(self, n_boot=100):
        """
        Bootstrap the input alignment.

        Args:
            n_boot (int): The number of bootstrap replicates to use.
        """
        self.bootstrapped_alignments = []
        for _ in range(n_boot):
            bootstrapped_alignment = self.alignment.copy()
            bootstrapped_alignment.loc[:, :] = np.random.choice(
                self.alignment.values.flatten(),
                size=self.alignment.shape,
                replace=True,
            ).reshape(self.alignment.shape)
            self.bootstrapped_alignments.append(bootstrapped_alignment)

    def calculate_z_scores_and_p_values(self, results):
        """
        Calculate Z-scores and P-values for the D-statistic methods.

        Args:
            results (list): A list of tuples containing the D-statistic values and their corresponding
                            standard errors for each permutation.

        Returns:
            z_scores (list): A list of Z-scores for each permutation.
            p_values (list): A list of P-values for each permutation.
            corrected_p_values (list): A list of Bonferroni-corrected P-values for each permutation.
        """
        d_statistics = [result[0] for result in results]
        standard_errors = [result[1] for result in results]

        # Calculate Z-scores
        z_scores = np.array(d_statistics) / np.array(standard_errors)

        # Calculate P-values
        p_values = 2 * norm.sf(np.abs(z_scores))

        # Apply Bonferroni correction
        corrected_p_values = np.minimum(p_values * len(p_values), 1)

        return z_scores, p_values, corrected_p_values

    def calculate_dfoil_statistic(self, p1, p2, p3, p4, outgroup):
        """
        Calculate the D-foil statistic for the given samples.

        Args:
            p1, p2, p3, p4 (str): Sample IDs.
            outgroup (str): Outgroup sample ID.

        Returns:
            float: D-foil statistic.
        """
        site_patterns = self.calculate_site_patterns(p1, p2, p3, p4, outgroup)

        site_patterns_array = np.array(
            [
                site_patterns["ABBA"],
                site_patterns["BABA"],
                site_patterns["BBAA"],
                site_patterns["AAAB"],
                site_patterns["AABB"],
            ]
        )

        dfoil_stat = (site_patterns_array[0] - site_patterns_array[1]) / (
            site_patterns_array[0]
            + site_patterns_array[1]
            + site_patterns_array[2]
            + site_patterns_array[3]
            + site_patterns_array[4]
        )

        return dfoil_stat

    def perform_permutations(
        self,
        sample_combinations,
        test_type="patterson",
        n_boot=1000,
        n_processes=None,
    ):
        """
        Perform permutations for the specified test_type using the loaded alignment.

        Args:
            sample_combinations (list(tuple())): List of tuples containing all sample combinations for the Dtest.
            test_type (str): The test type to run ("patterson", "partitioned", or "dfoil").
            n_boot (int): The number of bootstrap replicates to use.
            n_processes (int): The number of processes to use for multiprocessing.

        Returns:
            results_df (pd.DataFrame): A DataFrame containing the D-statistic, Z-score, and corrected P-value
                                       for each permutation.
        """
        # Create a partial function with the fixed test_type and n_boot arguments
        run_permutations_partial = partial(
            self._run_permutations, test_type, n_boot
        )

        if n_processes is None or n_processes <= 1:
            # Run permutations sequentially
            results_df = pd.concat(
                map(run_permutations_partial, sample_combinations)
            )
        else:
            # Run permutations in parallel
            with multiprocessing.Pool(n_processes) as p:
                results_list = p.map(
                    run_permutations_partial, sample_combinations
                )
                results_df = pd.concat(results_list)

        results_df.reset_index(drop=True, inplace=True)

        return results_df

    def _run_permutations(self, test_type, n_boot, samples):
        """
        Run a single permutation for the specified test_type using the given samples and
        the number of bootstrap replicates.

        Args:
            test_type (str): The test type to run ("patterson" or "partitioned").
            n_boot (int): The number of bootstrap replicates to use.
            samples (tuple): A tuple containing the samples for the permutation.

        Returns:
            results_df (pd.DataFrame): A DataFrame containing the D-statistic, Z-score, and corrected P-value
                                       for the permutation.
        """

        self.bootstrap_alignment(n_boot)
        boot_results = []

        for bootstrapped_alignment in self.bootstrapped_alignments:
            if test_type == "patterson":
                (
                    d_statistic,
                    standard_error,
                ) = self.patterson_d_statistic_permutation(
                    *samples, alignment=bootstrapped_alignment
                )
            elif test_type == "partitioned":
                (
                    d_statistic,
                    standard_error,
                ) = self.partitioned_d_statistic_permutation(
                    *samples, alignment=bootstrapped_alignment
                )
            elif test_type == "dfoil":
                d_statistic, standard_error = self.calculate_dfoil_statistic(
                    *samples, alignment=bootstrapped_alignment
                )
            else:
                raise ValueError(
                    "Invalid test_type. Must be 'patterson', 'partitioned', or 'dfoil'."
                )

            boot_results.append((d_statistic, standard_error))

        (
            z_scores,
            p_values,
            corrected_p_values,
        ) = self.calculate_z_scores_and_p_values(boot_results)

        # Store the results for this permutation
        results = {
            "Samples": str(samples),
            "Z-score": np.mean(z_scores),
            "P-value": np.mean(p_values),
            "Corrected P-value": np.mean(corrected_p_values),
        }

        # Convert results to a pandas DataFrame
        results_df = pd.DataFrame([results])

        return results_df

    # def get_population_sequences(self, population_name):
    #     """
    #     Retrieve sequences of the samples belonging to a given population.

    #     Args:
    #         population_name (str): The name of the population to retrieve sequences for.

    #     Returns:
    #         list: A list of sequences for the samples in the specified population.
    #     """
    #     # Assuming you have a dictionary that maps population names to sample names

    #     sample_names = self.populations[population_name]
    #     population_sequences = []
    #     for record in self.alignment:
    #         if record.id in sample_names:
    #             population_sequences.append(record)

    #     return population_sequences

    def calculate_1d_sfs(self, population, filter_singletons=False):
        pop_seqs = self.nremover.get_population_sequences(population)
        num_samples = len(pop_seqs)

        allele_counts = [
            Counter(
                [
                    seq[pos]
                    for seq in pop_seqs
                    if seq[pos] != "-" and seq[pos] != "N"
                ]
            )
            for pos in range(len(pop_seqs[0]))
        ]

        sfs = np.zeros(num_samples)
        for ac in allele_counts:
            if len(ac) == 2:  # biallelic
                count = min(ac.values())
                sfs[count - 1] += 1

        if filter_singletons:
            sfs = self.nremover.filter_singletons_sfs(sfs)

        return sfs

    def calculate_2d_sfs(
        self, population1, population2, filter_singletons=False
    ):
        pop1_seqs = self.nremover.get_population_sequences(population1)
        pop2_seqs = self.nremover.get_population_sequences(population2)
        num_samples1 = len(pop1_seqs)
        num_samples2 = len(pop2_seqs)

        allele_counts1 = [
            Counter(
                [
                    seq[pos]
                    for seq in pop1_seqs
                    if seq[pos] != "-" and seq[pos] != "N"
                ]
            )
            for pos in range(len(pop1_seqs[0]))
        ]

        allele_counts2 = [
            Counter(
                [
                    seq[pos]
                    for seq in pop2_seqs
                    if seq[pos] != "-" and seq[pos] != "N"
                ]
            )
            for pos in range(len(pop2_seqs[0]))
        ]

        sfs2d = np.zeros((num_samples1, num_samples2))
        for ac1, ac2 in zip(allele_counts1, allele_counts2):
            if len(ac1) == 2 and len(ac2) == 2:  # biallelic
                count1 = min(ac1.values())
                count2 = min(ac2.values())
                sfs2d[count1 - 1, count2 - 1] += 1

        if filter_singletons:
            sfs2d = self.nremover.filter_singletons_sfs(sfs2d)

        return sfs2d

    def detect_fst_outliers(self, fst_threshold):
        """
        Detect Fst outliers from the SNP data.

        Args:
            fst_threshold (float): Fst threshold for detecting outliers.

        Returns:
            pd.DataFrame: A DataFrame containing the Fst outliers.
        """

        def calculate_fst(column):
            allele_counts = (
                column.groupby(self.popmap["PopulationID"])
                .value_counts()
                .unstack(fill_value=0)
            )
            total_allele_counts = allele_counts.sum(axis=1)

            p = allele_counts.div(total_allele_counts, axis=0)
            p_mean = p.mean(axis=0)

            # Calculate Fst using the Weir and Cockerham (1984) formula
            nc = (
                total_allele_counts
                - (total_allele_counts**2).sum() / total_allele_counts.sum()
            )
            a = (p**2).div(nc, axis=0).sum(axis=0) - p_mean**2
            b = (
                p.mul(1 - p, axis=0).div(nc, axis=0).sum(axis=0)
                - (1 - p_mean) * p_mean
            )

            fst = a / (a + b)
            return fst

        chi2_stat = -2 * np.log(1 - fst_values)
        p_values = 1 - chi2.cdf(chi2_stat, df=1)

        fst_values = self.alignment.apply(calculate_fst, axis=0)
        fst_outliers = fst_values[fst_values >= fst_threshold]
        p_outliers = p_values[fst_values >= fst_threshold]

        return pd.DataFrame({"Fst": fst_outliers, "P-value": p_outliers})

    def observed_heterozygosity(self):
        """
        Calculate observed heterozygosity (Ho) for each locus.

        Returns:
            pd.Series: A pandas Series containing the observed heterozygosity values.
        """
        ho = (self.alignment == 1).sum(axis=0) / self.alignment.shape[0]
        return ho

    def expected_heterozygosity(self):
        """
        Calculate expected heterozygosity (He) for each locus.

        Returns:
            pd.Series: A pandas Series containing the expected heterozygosity values.
        """
        allele_freqs = (
            self.alignment.replace({0: np.nan, 2: np.nan}).mean(axis=0) / 2
        )
        he = 2 * allele_freqs * (1 - allele_freqs)
        return he

    def nucleotide_diversity(self):
        """
        Calculate nucleotide diversity (Pi) for each locus.

        Returns:
            pd.Series: A pandas Series containing the nucleotide diversity values.
        """
        n = self.alignment.shape[0]
        he = self.expected_heterozygosity()
        pi = (n / (n - 1)) * he
        return pi

    def wrights_fst(self):
        """
        Calculate Wright's fixation index (Fst) for each locus.

        Returns:
            pd.Series: A pandas Series containing the Fst values.
        """
        ho = self.observed_heterozygosity()
        he = self.expected_heterozygosity()
        fst = (he - ho) / he
        return fst

    def summary_statistics(self, save_plots=True):
        """
        Calculate a suite of summary statistics for SNP data.

        Returns:
            pd.DataFrame: A DataFrame containing the calculated summary statistics.
        """
        ho = self.observed_heterozygosity()
        he = self.expected_heterozygosity()
        pi = self.nucleotide_diversity()
        fst = self.wrights_fst()

        summary_stats = pd.DataFrame(
            {"Ho": ho, "He": he, "Pi": pi, "Fst": fst}
        )

        if save_plots:
            self.plotting.plot_summary_statistics(summary_stats)

        return summary_stats

    def perform_amova(self):
        # Convert self.alignment to allel.HaplotypeArray format
        haplotype_array = allel.HaplotypeArray(self.alignment)

        # Create a list of population indices
        pop_indices = [
            np.where(np.array(self.sample_population_map) == pop_id)[0]
            for pop_id in self.population_sample_map.keys()
        ]

        # Perform AMOVA
        amova_results = allel.stats.distance.amova(
            haplotype_array, pop_indices
        )

        return amova_results

    def perform_samova(self, n_groups):
        # Convert self.alignment to allel.HaplotypeArray format
        haplotype_array = allel.HaplotypeArray(self.alignment)

        # Create a list of population indices
        pop_indices = [
            np.where(np.array(self.sample_population_map) == pop_id)[0]
            for pop_id in self.population_sample_map.keys()
        ]

        # Perform SAMOVA
        samova_results = allel.stats.distance.samova(
            haplotype_array, pop_indices, n_groups
        )

        return samova_results

    def perform_pca(
        self, n_components="auto", save_plot=True, plot_dimensions=2
    ):
        """
        Perform Principal Components Analysis (PCA) on self.alignment.

        Args:
            n_components (str, int, or float): The number of components to keep. If "auto" (default), KneeLocator will
            determine the optimal number of components based on the explained variance.

        Returns:
            sklearn.PCA: The fitted PCA object.
        """

        # Create a mapping dictionary
        char_to_int = {
            "A": 0,
            "T": 1,
            "C": 2,
            "G": 3,
            "N": np.nan,
            "-": np.nan,
            "R": 0,  # A/G, A is more frequent
            "Y": 2,  # C/T, C is more frequent
            "S": 2,  # G/C, C is more frequent
            "W": 0,  # A/T, A is more frequent
            "K": 1,  # G/T, T is more frequent
            "M": 0,  # A/C, A is more frequent
            "B": 2,  # C/G/T, C is most frequent
            "D": 0,  # A/G/T, A is most frequent
            "H": 0,  # A/C/T, A is most frequent
            "V": 0,  # A/C/G, A is most frequent
        }

        # Convert the input alignment to a numpy array of sequences
        alignment_array = np.array(
            [list(str(record.seq)) for record in self.alignment]
        )

        alignment_array = np.array(
            [[char_to_int[char] for char in row] for row in alignment_array]
        )

        imp = SimpleImputer(strategy="most_frequent")
        X = imp.fit_transform(alignment_array)

        ohe = OneHotEncoder(sparse_output=False)
        X = ohe.fit_transform(X)

        if n_components == "auto":
            pca = PCA()
            pca.fit(X)
            kneelocator = KneeLocator(
                range(1, len(pca.explained_variance_) + 1),
                pca.explained_variance_ratio_,
                curve="convex",
                direction="decreasing",
            )
            n_components = kneelocator.elbow

        pca = PCA(n_components=n_components)
        pca.fit(X)

        if save_plot:
            self.plotting.plot_pca(
                pca, X, self.popmap, dimensions=plot_dimensions
            )

        return pca

    def perform_dapc(
        self, n_components=None, save_plot=True, plot_dimensions=2
    ):
        """
        Perform Discriminant Analysis of Principal Components (DAPC) on self.alignment.

        Args:
            n_components (int, optional): The number of components to keep. If not provided, use cross-validation with
            the root mean squared error to select the optimal number of principal components.

        Returns:
            sklearn.discriminant_analysis.LinearDiscriminantAnalysis: The fitted DAPC object.
        """
        if n_components is None:
            pca = PCA()
            pca.fit(self.alignment)
            pca_transformed = pca.transform(self.alignment)

            min_rmse = float("inf")
            best_n_components = None

            for n in range(1, len(pca.explained_variance_) + 1):
                lda = LinearDiscriminantAnalysis(n_components=n)
                X_train, X_test, y_train, y_test = train_test_split(
                    pca_transformed[:, :n],
                    self.popmap["PopulationID"],
                    test_size=0.3,
                )
                lda.fit(X_train, y_train)
                y_pred = lda.predict(X_test)
                rmse = mean_squared_error(y_test, y_pred, squared=False)

                if rmse < min_rmse:
                    min_rmse = rmse
                    best_n_components = n

            n_components = best_n_components

        pca = PCA(n_components=n_components)
        pca_transformed = pca.fit_transform(self.alignment)
        dapc = LinearDiscriminantAnalysis()
        dapc.fit(pca_transformed, self.popmap["PopulationID"])

        if save_plot:
            self.plotting.plot_dapc(
                dapc, self.popmap, dimensions=plot_dimensions
            )

        return dapc

    def impute_genotypes(self, method="xgboost", tree_file=None):
        """
        Impute missing genotypes using the specified method.

        Args:
            method (str): Imputation method. Options are "phylogeny" or "xgboost". Default is "xgboost".
            tree_file (str): Path to a Newick or Nexus file containing the phylogenetic tree. Required if method="phylogeny".
        """
        if method == "phylogeny":
            self._impute_phylogeny(tree_file)
        elif method == "xgboost":
            self._impute_xgboost()
        else:
            raise ValueError(
                "Invalid imputation method. Choose 'phylogeny' or 'xgboost'."
            )

    def _impute_phylogeny(self, tree_file):
        if tree_file is None:
            raise ValueError(
                "tree_file is required when using the phylogeny imputation method."
            )

        # Load the tree
        tree = Tree(tree_file)

    def _impute_phylogeny(self, tree_file):
        if tree_file is None:
            raise ValueError(
                "tree_file is required when using the phylogeny imputation method."
            )

        # Load the tree
        tree = Tree(tree_file)

        # Impute missing genotypes using the phylogenetic tree
        for site in self.alignment.columns:
            for sample in self.alignment.index:
                if pd.isna(self.alignment.at[sample, site]):
                    # Find the nearest non-missing neighbor in the phylogenetic tree
                    node = tree.search_nodes(name=sample)[0]
                    while pd.isna(self.alignment.at[node.name, site]):
                        node = node.get_closest_leaf()[0]

                    # Impute the missing genotype with the genotype of the nearest non-missing neighbor
                    self.alignment.at[sample, site] = self.alignment.at[
                        node.name, site
                    ]



import pkg_resources


def get_installed_packages():
    installed_packages = [
        package.key
        for package in pkg_resources.working_set
        if package.location.endswith(
            ".dist-info"
        )  # Filter out individual modules
    ]
    return installed_packages


# Get the list of installed packages
packages = get_installed_packages()

# Print the installed packages
for package in packages:
    print(package)



import sys
import os
import functools
import psutil
import time
import subprocess
import re
import logging
import platform

import numpy as np
import pandas as pd


# from skopt import BayesSearchCV


def align_columns(rows, alignment="left"):
    """
    Aligns the columns of a list of tab-separated strings.

    Args:
        rows (list): A list of tab-separated strings.
        alignment (str): The type of alignment ("left", "right", "center").

    Returns:
        list: A list of aligned, tab-separated strings.
    """
    # Split each row into its columns
    split_rows = [row.split("\t") for row in rows]
    
    # Get the maximum width for each column
    max_widths = [max(len(str(item)) for item in col) for col in zip(*split_rows)]
    
    # Functions to align text
    align_func = {
        "left": str.ljust,
        "right": str.rjust,
        "center": str.center
    }
    
    # Align each column and join them back into strings
    aligned_rows = []
    for row in split_rows:
        aligned_row = [
            align_func[alignment](str(cell), width)
            for cell, width in zip(row, max_widths)
        ]
        aligned_rows.append("\t".join(aligned_row))
    
    return aligned_rows


def validate_input_type(X, return_type="array"):
    """
    Validates the input type and returns it as a specified type.

    This function checks if the input `X` is a pandas DataFrame, numpy array, or a list of lists. It then converts `X` to the specified `return_type` and returns it.

    Args:
        X (pandas.DataFrame, numpy.ndarray, or List[List[int]]): The input data to validate and convert.

        return_type (str, optional): The type of the returned object. Supported options include: "df" (DataFrame), "array" (numpy array), and "list". Defaults to "array".

    Returns:
        pandas.DataFrame, numpy.ndarray, or List[List[int]]: The input data converted to the desired return type.

    Raises:
        TypeError: If `X` is not of type pandas.DataFrame, numpy.ndarray, or List[List[int]].

        ValueError: If an unsupported `return_type` is provided. Supported types are "df", "array", and "list".

    Example:
        X = [[1, 2, 3], [4, 5, 6]]
        print(validate_input_type(X, "df"))  # Outputs: a DataFrame with the data from `X`.
    """
    if not isinstance(X, (pd.DataFrame, np.ndarray, list)):
        raise TypeError(
            f"X must be of type pandas.DataFrame, numpy.ndarray, "
            f"or List[List[int]], but got {type(X)}"
        )

    if return_type == "array":
        if isinstance(X, pd.DataFrame):
            return X.to_numpy()
        elif isinstance(X, list):
            return np.array(X)
        elif isinstance(X, np.ndarray):
            return X.copy()

    elif return_type == "df":
        if isinstance(X, pd.DataFrame):
            return X.copy()
        elif isinstance(X, (np.ndarray, list)):
            return pd.DataFrame(X)

    elif return_type == "list":
        if isinstance(X, list):
            return X
        elif isinstance(X, np.ndarray):
            return X.tolist()
        elif isinstance(X, pd.DataFrame):
            return X.values.tolist()

    else:
        raise ValueError(
            f"Unsupported return type provided: {return_type}. Supported types "
            f"are 'df', 'array', and 'list'"
        )


def generate_random_dataset(
    min_value=0,
    max_value=2,
    nrows=35,
    ncols=20,
    min_missing_rate=0.15,
    max_missing_rate=0.5,
):
    """
    Generate a random integer dataset that can be used for testing.

    This function generates a 2D numpy array of random integers between `min_value` and `max_value` (inclusive). It also adds randomly missing values of random proportions between `min_missing_rate` and `max_missing_rate`.

    Args:
        min_value (int, optional): Minimum value to use. Defaults to 0.
        max_value (int, optional): Maximum value to use. Defaults to 2.
        nrows (int, optional): Number of rows to use. Defaults to 35.
        ncols (int, optional): Number of columns to use. Defaults to 20.
        min_missing_rate (float, optional): Minimum proportion of missing data per column. Defaults to 0.15.
        max_missing_rate (float, optional): Maximum proportion of missing data per column. Defaults to 0.5.

    Returns:
        numpy.ndarray: A 2D numpy array of shape (nrows, ncols) containing the randomly generated dataset.

    Raises:
        AssertionError: If any of the input parameters are out of their expected ranges.

    Example:
        print(generate_random_dataset(0, 5, 5, 5, 0.1, 0.3))
        # Outputs: a 5x5 numpy array with random integers between 0 and 5 and some missing values.
    """
    assert (
        min_missing_rate >= 0 and min_missing_rate < 1.0
    ), f"min_missing_rate must be >= 0 and < 1.0, but got {min_missing_rate}"

    assert (
        max_missing_rate > 0 and max_missing_rate < 1.0
    ), f"max_missing_rate must be > 0 and < 1.0, but got {max_missing_rate}"

    assert nrows > 1, f"nrows must be > 1, but got {nrows}"
    assert ncols > 1, f"ncols must be > 1, but got {ncols}"

    try:
        min_missing_rate = float(min_missing_rate)
        max_missing_rate = float(max_missing_rate)
    except TypeError:
        sys.exit(
            "min_missing_rate and max_missing_rate must be of type float or "
            "must be cast-able to type float"
        )

    X = np.random.randint(
        min_value, max_value + 1, size=(nrows, ncols)
    ).astype(float)
    for i in range(X.shape[1]):
        drop_rate = int(
            np.random.choice(
                np.arange(min_missing_rate, max_missing_rate, 0.02), 1
            )[0]
            * X.shape[0]
        )

        rows = np.random.choice(np.arange(0, X.shape[0]), size=drop_rate)
        X[rows, i] = np.nan

    return X


def generate_012_genotypes(
    nrows=35,
    ncols=20,
    max_missing_rate=0.5,
    min_het_rate=0.001,
    max_het_rate=0.3,
    min_alt_rate=0.001,
    max_alt_rate=0.3,
):
    """
    Generates a 2D numpy array of random 012-encoded genotypes.

    Allows users to control the rate of reference (0's), heterozygote (1's), and alternate alleles (2's). Will insert a random proportion between `min_het_rate` and `max_het_rate` and `min_alt_rate` and `max_alt_rate` and from no missing data to a proportion of `max_missing_rate`.

    Args:
        nrows (int, optional): Number of rows to generate. Defaults to 35.

        ncols (int, optional): Number of columns to generate. Defaults to 20.

        max_missing_rate (float, optional): Maximum proportion of missing data to use. Defaults to 0.5.

        min_het_rate (float, optional): Minimum proportion of heterozygotes (1's) to insert. Defaults to 0.001.

        max_het_rate (float, optional): Maximum proportion of heterozygotes (1's) to insert. Defaults to 0.3.

        min_alt_rate (float, optional): Minimum proportion of alternate alleles (2's) to insert. Defaults to 0.001.

        max_alt_rate (float, optional): Maximum proportion of alternate alleles (2's) to insert. Defaults to 0.3.

    Returns:
        numpy.ndarray: A 2D numpy array of shape (nrows, ncols) containing the generated 012-encoded genotypes.

    Raises:
        AssertionError: If any of the input parameters are out of their expected ranges.

    Example:
        print(generate_012_genotypes(5, 5, 0.2, 0.1, 0.3, 0.1, 0.3))
        # Outputs: a 5x5 numpy array with 012-encoded genotypes.
    """
    assert (
        min_het_rate > 0 and min_het_rate <= 1.0
    ), f"min_het_rate must be > 0 and <= 1.0, but got {min_het_rate}"

    assert (
        max_het_rate > 0 and max_het_rate <= 1.0
    ), f"max_het_rate must be > 0 and <= 1.0, but got {max_het_rate}"

    assert (
        min_alt_rate > 0 and min_alt_rate <= 1.0
    ), f"min_alt_rate must be > 0 and <= 1.0, but got {min_alt_rate}"

    assert (
        max_alt_rate > 0 and max_alt_rate <= 1.0
    ), f"max_alt_rate must be > 0 and <= 1.0, but got {max_alt_rate}"

    assert nrows > 1, f"The number of rows must be > 1, but got {nrows}"

    assert ncols > 1, f"The number of columns must be > 1, but got {ncols}"

    assert (
        max_missing_rate > 0 and max_missing_rate < 1.0
    ), f"max_missing rate must be > 0 and < 1.0, but got {max_missing_rate}"

    try:
        min_het_rate = float(min_het_rate)
        max_het_rate = float(max_het_rate)
        min_alt_rate = float(min_alt_rate)
        max_alt_rate = float(max_alt_rate)
        max_missing_rate = float(max_missing_rate)
    except TypeError:
        sys.exit(
            "max_missing_rate, min_het_rate, max_het_rate, min_alt_rate, and "
            "max_alt_rate must be of type float, or must be cast-able to type "
            "float"
        )

    X = np.zeros((nrows, ncols))
    for i in range(X.shape[1]):
        het_rate = int(
            np.ceil(
                np.random.choice(
                    np.arange(min_het_rate, max_het_rate, 0.02), 1
                )[0]
                * X.shape[0]
            )
        )

        alt_rate = int(
            np.ceil(
                np.random.choice(
                    np.arange(min_alt_rate, max_alt_rate, 0.02), 1
                )[0]
                * X.shape[0]
            )
        )

        het = np.sort(
            np.random.choice(
                np.arange(0, X.shape[0]), size=het_rate, replace=False
            )
        )

        alt = np.sort(
            np.random.choice(
                np.arange(0, X.shape[0]), size=alt_rate, replace=False
            )
        )

        sidx = alt.argsort()
        idx = np.searchsorted(alt, het, sorter=sidx)
        idx[idx == len(alt)] = 0
        het_unique = het[alt[sidx[idx]] != het]

        X[alt, i] = 2
        X[het_unique, i] = 1

        drop_rate = int(
            np.random.choice(np.arange(0.15, max_missing_rate, 0.02), 1)[0]
            * X.shape[0]
        )

        missing = np.random.choice(np.arange(0, X.shape[0]), size=drop_rate)

        X[missing, i] = np.nan

    print(
        f"Created a dataset of shape {X.shape} with {np.isnan(X).sum()} total missing values"
    )

    return X


def unique2D_subarray(a):
    """
    Returns unique subarrays for each column from a 2D numpy array.

    Args:
        a (numpy.ndarray): The 2D numpy array to process.

    Returns:
        numpy.ndarray: A 2D numpy array containing only the unique subarrays from the input array.

    Example:
        a = np.array([[1, 2, 3], [1, 2, 3], [4, 5, 6]])
        print(unique2D_subarray(a))  # Outputs: [[1, 2, 3], [4, 5, 6]]

    Note:
        The function first views the input array as a 1D array with a custom data type, then uses numpy's unique function to find the unique elements.
    """
    dtype1 = np.dtype((np.void, a.dtype.itemsize * np.prod(a.shape[1:])))
    b = np.ascontiguousarray(a.reshape(a.shape[0], -1)).view(dtype1)
    return a[np.unique(b, return_index=1, axis=-1)[1]]


def get_indices(l):
    """
    Takes a list and returns a dictionary where keys are the unique elements in the list and values are lists of indices where each element appears.

    Args:
        l (List[Any]): The list to process.

    Returns:
        dict: A dictionary where keys are the unique elements in the list and values are lists of indices where each element appears.

    Example:
        print(get_indices([0, 1, 1, 0, 0]))  # Outputs: {0: [0, 3, 4], 1: [1, 2]}

    Note:
        The function uses a set to get the unique elements in the list, and then iterates over the list to get the indices.
    """
    ret = dict()
    for member in set(l):
        ret[member] = list()
    i = 0
    for el in l:
        ret[el].append(i)
        i += 1
    return ret


def all_zero(l):
    """
    Checks whether a list consists of all zeros.

    This function returns True if the supplied list contains only zeros (integer, float, or string representations).
    It returns False if the list contains any non-zero values or if the list is empty.

    Args:
        l (List[Union[int, float, str]]): The list to check.

    Returns:
        bool: True if all elements in the list are zeros, False otherwise.

    Example:
        print(all_zero([0, 0.0, '0', '0.0']))  # Outputs: True
        print(all_zero([0, 1, 2]))  # Outputs: False
        print(all_zero([]))  # Outputs: False
    """
    values = set(l)
    if len(values) > 1:
        return False
    elif len(values) == 1 and l[0] in [0, 0.0, "0", "0.0"]:
        return True
    else:
        return False


def weighted_draw(d, num_samples=1):
    """
    Draws samples from a dictionary where keys are choices and values are their corresponding weights.

    Args:
        d (dict): The dictionary from which to draw samples. Keys are the choices and values are the corresponding weights.
        num_samples (int, optional): The number of samples to draw. Defaults to 1.

    Returns:
        numpy.ndarray: An array of drawn samples.

    Example:
        d = {'a': 0.5, 'b': 0.3, 'c': 0.2}
        print(weighted_draw(d, 10))  # Outputs: array(['a', 'b', 'a', 'a', 'b', 'a', 'c', 'a', 'a', 'b'])

    Note:
        The function uses numpy's random.choice function for drawing samples.
    """
    choices = list(d.keys())
    weights = list(d.values())
    return np.random.choice(choices, num_samples, p=weights)


def get_attributes(cls):
    """
    Retrieves the attributes of a class or an instance.

    Args:
        cls (object): The class or instance from which to retrieve attributes.

    Returns:
        dict: A dictionary where the keys are the attribute names and the values are the attribute values. Only includes attributes that do not start with "__" and are not callable.

    Example:
        class MyClass:
            x = 1
            y = 2

        print(get_attributes(MyClass))  # Outputs: {'x': 1, 'y': 2}
    """
    return {
        k: v
        for k, v in cls.__dict__.items()
        if not k.startswith("__") and not callable(v)
    }


def measure_performance_for_instance_method(func):
    """
    Decorator for measuring the performance of an instance method. The performance metrics include CPU load, memory footprint, and execution time.

    Args:
        func (callable): The instance method to be measured.

    Returns:
        callable: The decorated instance method that measures its performance when called.

    Note:
        The performance data is stored in the `resource_data` attribute of the instance, under a key with the name of the method. The data is a dictionary with keys 'cpu_load', 'memory_footprint', and 'execution_time'.
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        # Create nested dictionary for function's resource data
        self.resource_data[func.__name__] = {}

        # Measure CPU load
        cpu_load = psutil.cpu_percent()
        self.resource_data[func.__name__]["cpu_load"] = cpu_load

        # Measure memory footprint
        process = psutil.Process(os.getpid())
        memory_footprint = process.memory_info().rss
        memory_footprint_mb = memory_footprint / (
            1024 * 1024
        )  # Convert bytes to megabytes
        self.resource_data[func.__name__][
            "memory_footprint"
        ] = memory_footprint_mb

        # Measure execution time
        start_time = time.time()
        result = func(self, *args, **kwargs)
        end_time = time.time()
        execution_time = end_time - start_time
        self.resource_data[func.__name__]["execution_time"] = execution_time
        return result

    return wrapper


def measure_performance_for_class_method(func):
    """
    Decorator for measuring the performance of a class method. The performance metrics include CPU load, memory footprint, and execution time.

    Args:
        func (callable): The class method to be measured.

    Returns:
        callable: The decorated class method that measures its performance when called.
    """

    @functools.wraps(func)
    def wrapper(cls, *args, **kwargs):
        # Create nested dictionary for function's resource data
        cls.resource_data[func.__name__] = {}

        # Measure CPU load
        cpu_load = psutil.cpu_percent()
        cls.resource_data[func.__name__]["cpu_load"] = cpu_load

        # Measure memory footprint
        process = psutil.Process(os.getpid())
        memory_footprint = process.memory_info().rss
        memory_footprint_mb = memory_footprint / (
            1024 * 1024
        )  # Convert bytes to megabytes
        cls.resource_data[func.__name__][
            "memory_footprint"
        ] = memory_footprint_mb

        # Measure execution time
        start_time = time.time()
        result = func(cls, *args, **kwargs)
        end_time = time.time()
        execution_time = end_time - start_time
        cls.resource_data[func.__name__]["execution_time"] = execution_time
        return result

    return wrapper


def class_performance_decorator(measure=True):
    """
    Decorator for applying performance measurement to all callable attributes of a class that do not start with an underscore. 
    
    The performance metrics include CPU load, memory footprint, and execution time.

    Args:
        measure (bool, optional): If True, apply performance measurement. If False, return the class unchanged. Defaults to True.

    Returns:
        callable: The decorated class with performance measurement applied to its callable attributes, or the original class if measure is False.
    """

    def measure_decorator(cls):
        for attr_name, attr_value in cls.__dict__.items():
            if callable(attr_value) and not attr_name.startswith("_"):
                if isinstance(attr_value, (staticmethod, classmethod)):
                    decorated_func = measure_performance_for_class_method(
                        attr_value.__func__
                    )
                    setattr(cls, attr_name, classmethod(decorated_func))
                else:
                    decorated_func = measure_performance_for_instance_method(
                        attr_value
                    )
                    setattr(cls, attr_name, decorated_func)
        return cls

    return measure_decorator if measure else lambda cls: cls


def progressbar(it, prefix="", size=60, f=sys.stdout):
    """
    Generator that prints a progress bar to the console for an iterable.

    Args:
        it (iterable): The iterable to iterate over.

        prefix (str, optional): The prefix string to be printed before the progress bar. Defaults to an empty string.

        size (int, optional): The total width of the progress bar in characters. Defaults to 60.

        f (file-like object, optional): The file-like object to which the progress bar is printed. Defaults to sys.stdout.

    Yields:
        The next item from the iterable.

    Note:
        The progress bar is printed in the format: "{prefix}[{# * progress}{. * (size - progress)}] {progress}/{total}"
    """
    count = len(it)

    def show(j):
        x = int(size * j / count)
        f.write(
            "%s[%s%s] %i/%i\r" % (prefix, "#" * x, "." * (size - x), j, count)
        )
        f.flush()

    show(0)
    for i, item in enumerate(it):
        yield item
        show(i + 1)
    f.write("\n")
    f.flush()


def get_processor_name():
    """
    Retrieves the name of the processor of the system.

    Returns:
        str: The name of the processor. If the system is not recognized (not Windows, Darwin, or Linux), an empty string is returned.

    Note:
        For Windows, it uses the platform.processor() function.

        For Darwin (Mac OS), it returns 'Intel' if the architecture starts with 'i', otherwise it returns the architecture.

        For Linux, it reads from /proc/cpuinfo to get the model name of the processor.
    """
    if platform.system() == "Windows":
        return platform.processor()
    elif platform.system() == "Darwin":
        # os.environ["PATH"] = os.environ["PATH"] + os.pathsep + "/usr/sbin"
        arch = platform.processor()
        if arch[0] == "i":
            return "Intel"
        else:
            return arch
    elif platform.system() == "Linux":
        command = "cat /proc/cpuinfo"
        all_info = subprocess.check_output(command, shell=True).strip()
        all_info = all_info.decode("utf-8")
        for line in all_info.split("\n"):
            if "model name" in line:
                return re.sub(".*model name.*:", "", line, 1)
    return ""


class HiddenPrints:
    """
    Context manager for suppressing print statements within its scope.

    This class redirects stdout to os.devnull within its scope, effectively suppressing all print statements. When the scope is exited, stdout is restored to its original state.

    Example:
        with HiddenPrints():
            print("This will not be printed.")

    Attributes:
        _original_stdout (file-like object): The original stdout stream, stored before it is redirected.

    Note:
        This class does not suppress output from sys.stderr.
    """

    def __enter__(self):
        """
        Enter the runtime context. Redirects stdout to os.devnull.

        Returns:
            self
        """
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit the runtime context. Restores stdout to its original state.

        Args:
            exc_type (Type[BaseException], optional): The type of the exception that caused the context to be exited, if any.
            exc_val (BaseException, optional): The instance of the exception that caused the context to be exited, if any.
            exc_tb (traceback, optional): A traceback object encapsulating the call stack at the point where the exception was raised, if any.

        Returns:
            None
        """
        sys.stdout.close()
        sys.stdout = self._original_stdout


class StreamToLogger(object):
    """
    File-like stream object that redirects writes to a logger instance.

    This class is designed to redirect stdout or stderr to a logging object. It has a write method that splits input on newlines and logs each line separately, and a flush method that logs any remaining input.

    Attributes:
        logger (logging.Logger): The logger instance to which writes are redirected.
        log_level (int, optional): The log level at which messages are logged. Defaults to logging.INFO.
        linebuf (str): A buffer for storing partial lines.

    Example:
        logger = logging.getLogger('my_logger')
        sys.stdout = StreamToLogger(logger, logging.INFO)

    Note:
        This class does not close the logger when it is garbage collected.
    """

    def __init__(self, logger, log_level=logging.INFO):
        """
        Initialize the StreamToLogger instance.

        Args:
            logger (logging.Logger): The logger instance to which writes are redirected.
            log_level (int, optional): The log level at which messages are logged. Defaults to logging.INFO.
        """
        self.logger = logger
        self.log_level = log_level
        self.linebuf = ""

    def write(self, buf):
        """
        Write the specified string to the stream.

        Args:
            buf (str): The string to write to the stream.
        """
        temp_linebuf = self.linebuf + buf
        self.linebuf = ""
        for line in temp_linebuf.splitlines(True):
            # From the io.TextIOWrapper docs:
            #   On output, if newline is None, any '\n' characters written
            #   are translated to the system default line separator.
            # By default sys.stdout.write() expects '\n' newlines and then
            # translates them so this is still cross platform.
            if line[-1] == "\n":
                self.logger.log(self.log_level, line.rstrip())
            else:
                self.linebuf += line

    def flush(self):
        """
        Flush the stream.

        If the stream buffer is not empty, logs its contents and then clears it.
        """
        if self.linebuf != "":
            self.logger.log(self.log_level, self.linebuf.rstrip())
        self.linebuf = ""



import sys
import os

sys.path.append(os.path.normpath(os.getcwd()))
from snpio.read_input.genotype_data import GenotypeData

# Instantiate GenotypeData object
genotype_data = GenotypeData(
    filename="example_data/vcf_files/phylogen_subset14K.vcf.gz",
    force_popmap=True,
    filetype="auto",
    popmapfile="example_data/popmaps/phylogen_nomx.popmap",
    guidetree="example_data/trees/test.tre",
    siterates_iqtree="example_data/trees/test14K.rate",
    qmatrix_iqtree="example_data/trees/test.iqtree",
)

# Access basic properties
num_snps = genotype_data.num_snps
num_inds = genotype_data.num_inds
populations = genotype_data.populations
popmap = genotype_data.popmap
genotype_data.popmap_inverse
samples = genotype_data.samples
inputs = genotype_data.inputs
ref = genotype_data.ref
alt = genotype_data.alt
loci_indices = genotype_data.loci_indices
sample_indices = genotype_data.sample_indices
snp_data = genotype_data.snp_data

# Access other transformed data
genotypes_onehot = genotype_data.genotypes_onehot
genotypes_int = genotype_data.genotypes_int
alignment = genotype_data.alignment

# Access VCF file attributes
vcf_attributes = genotype_data.vcf_attributes

# Access additional properties
q_matrix = genotype_data.q
site_rates = genotype_data.site_rates
newick_tree = genotype_data.tree

genotype_data.plot_performance()



import sys
import os
import tempfile
import warnings
import pandas as pd
import numpy as np
from Bio.Align import MultipleSeqAlignment
from Bio import SeqUtils
from copy import deepcopy
from pathlib import Path

from snpio.plotting.plotting import Plotting
from snpio.read_input.genotype_data import GenotypeData


class NRemover2:
    """
    A class for filtering alignments based on the proportion of missing data in a genetic alignment.

    The class can filter out sequences (samples) and loci (columns) that exceed a missing data threshold.

    The loci can be filtered by global missing data proportions or if any given population exceeds the missing data threshold.

    A number of informative plots are also generated.

    Note:
        NRemover2 handles the following characters as missing data:
            - 'N'
            - '-'
            - '?'
            - '.'

        Thus, it treats gaps as missing data. Please keep this in mind when using NRemover2.

    Args:
        popgenio (GenotypeData): An instance of the GenotypeData class containing the genetic data alignment, population map, and populations.

    Attributes:
        alignment (list of Bio.SeqRecord.SeqRecord): The input alignment to filter.

        populations (list of str): The population for each sequence in the alignment.

    Properties:
        alignment (property): Property for accessing and setting the alignment.

        msa (property): Property for accessing and setting the msa.

        population_sequences (property): Property for accessing the sequences for each population.

    Methods:
        nremover: Runs the whole NRemover2 pipeline.

        filter_missing: Filters out sequences from the alignment that have more than a given proportion of missing data.

        filter_missing_pop: Filters out sequences from the alignment that have more than a given proportion of missing data in a specific population.

        filter_missing_sample: Filters out samples from the alignment that have more than a given proportion of missing data.

        filter_minor_allele_frequency: Filters out loci (columns) where the minor allele frequency is below the threshold.

        filter_monomorphic: Filters out monomorphic sites.

        filter_singletons: Filters out loci (columns) where the only variant is a singleton.

        filter_non_biallelic: Filter out loci (columns) that have more than 2 alleles.

        get_population_sequences: Returns the sequences for a specific population.

        count_iupac_alleles: Counts the number of occurrences of each IUPAC ambiguity code in a given column.

        count_unique_bases: Counts the number of unique bases in a given column.

        plot_missing_data_thresholds: Plots the proportion of missing data against the filtering thresholds.

        plot_sankey_filtering_report: Makes a Sankey plot showing the number of loci removed at each filtering step.

        print_filtering_report: Prints a summary of the filtering results.
    """

    def __init__(self, popgenio):
        self._msa = popgenio.alignment
        self._alignment = deepcopy(self._msa)
        self.popgenio = popgenio
        self.popmap = popgenio.popmap
        self.popmap_inverse = popgenio.popmap_inverse
        self.populations = self.popmap_inverse.keys()
        self.samples = popgenio.samples
        self.poplist = popgenio.populations
        self.prefix = popgenio.prefix

        self.loci_indices = None
        self.sample_indices = None

    def nremover(
        self,
        max_missing_global=1.0,
        max_missing_pop=1.0,
        max_missing_sample=1.0,
        min_maf=0.0,
        biallelic=False,
        monomorphic=False,
        singletons=False,
        search_thresholds=True,
        plot_dir_prefix="snpio",
        file_prefix=None,
        plot_format="png",
        dpi=300,
        suppress_cletus=True,
        included_steps=None,
    ):
        """
        Runs the NRemover2 pipeline for filtering alignments based on missing data, minor allele frequency, and monomorphic, non-biallelic, and singleton sites.

        Args:
            max_missing_global (float, optional): The maximum proportion of missing data allowed globally (across all samples). Defaults to 1.0.

            max_missing_pop (float, optional): The maximum proportion of missing data allowed within a population. Defaults to 1.0.

            max_missing_sample (float, optional): The maximum proportion of missing data allowed for a single sample. Defaults to 1.0.

            min_maf (float, optional): The minimum minor allele frequency threshold. Defaults to 0.0.

            biallelic (bool, optional): Whether to filter out non-biallelic loci. Defaults to False.

            monomorphic (bool, optional): Whether to filter out monomorphic loci. Defaults to False.

            singletons (bool, optional): Whether to filter out loci where the only variant is a singleton. Defaults to False.

            search_thresholds (bool, optional): Whether to search across multiple thresholds and make a plot for visualization. Defaults to True.

            plot_dir_prefix (str, optional): The prefix for the output plot directory. Defaults to "snpio".

            file_prefix (str, optional): The prefix for the output filename. If ``file_prefix`` is None, then no prefix is prepended to the filename. Defaults to None.

            plot_format (str, optional): Format to save plot to. Supported image formats include: "pdf", "svg", "png", and "jpeg" (or "jpg"). Defaults to "png".

            dpi (int, optional): DPI resolution of output plots. Defaults to 300.

            included_steps (list, optional): The steps to include in the Sankey plot. If None, all steps will be included. Defaults to None.

        Returns:
            GenotypeData: A GenotypeData object containing the filtered alignment, retained loci indices, and retained sample indices.

        """
        if not suppress_cletus:
            self.print_cletus()

        self.alignment = self.msa[:]

        aln_before = deepcopy(self.alignment)
        indices_loci_before = range(len(aln_before[0]))

        plot_dir = os.path.join(f"{self.prefix}_output", "nremover", "plots")
        Path(plot_dir).mkdir(exist_ok=True, parents=True)

        Plotting.plot_gt_distribution(
            self.popgenio.genotypes_int,
            plot_dir_prefix=plot_dir_prefix,
            file_prefix=file_prefix,
            plot_format=plot_format,
            dpi=dpi,
        )

        output_file = "missingness_threshold_search.png"

        if search_thresholds:
            self.search_thresholds_ = True
            self.plot_missing_data_thresholds(
                output_file,
                plot_dir_prefix=plot_dir_prefix,
                file_prefix=file_prefix,
                plot_format=plot_format,
                dpi=dpi,
            )

        steps = [
            (
                "Filter missing data (sample)",
                max_missing_sample < 1.0,
                max_missing_sample,
                self.filter_missing_sample,
                5,
            ),
            (
                "Filter monomorphic sites",
                monomorphic,
                monomorphic,
                self.filter_monomorphic,
                0,
            ),
            (
                "Filter singletons",
                singletons,
                singletons,
                self.filter_singletons,
                1,
            ),
            (
                "Filter non-biallelic sites",
                biallelic,
                biallelic,
                self.filter_non_biallelic,
                2,
            ),
            (
                "Filter missing data (global)",
                max_missing_global < 1.0,
                max_missing_global,
                self.filter_missing,
                3,
            ),
            (
                "Filter missing data (population)",
                max_missing_pop < 1.0,
                max_missing_pop,
                self.filter_missing_pop,
                4,
            ),
            (
                "Filter minor allele frequency",
                min_maf > 0.0,
                min_maf,
                self.filter_minor_allele_frequency,
                6,
            ),
        ]

        indices_loci_before = list(range(len(aln_before[0])))
        loci_removed_per_step = []
        filter_idx_dict = {}
        retained_indices = indices_loci_before

        for name, condition, threshold, filter_func, step_idx in steps:
            if condition:
                filtered_alignment, indices = filter_func(
                    threshold, alignment=self.alignment
                )
                loci_removed = len(self.alignment[0]) - len(filtered_alignment[0])

                if name != "Filter missing data (sample)":
                    loci_removed_per_step.append((name, loci_removed))

                    # Update retained_indices to only include the indices that were retained
                    retained_indices = [retained_indices[i] for i in indices]
                    filter_idx_dict[step_idx] = retained_indices
                else:
                    self.sample_indices = indices
                    self.samples = [self.samples[i] for i in indices]
                self.alignment = filtered_alignment
            else:
                loci_removed_per_step.append((name, 0))

        aln_after = deepcopy(self.alignment)

        max_step_idx = max(filter_idx_dict)
        self.loci_indices = filter_idx_dict[max_step_idx]

        self.print_filtering_report(aln_before, aln_after, loci_removed_per_step)

        if included_steps is None:
            included_steps = [
                step_idx for _, condition, _, _, step_idx in steps if condition
            ]

        plot_format = plot_format.lower()
        outfile = "sankey_filtering_report.html"

        Plotting.plot_sankey_filtering_report(
            loci_removed_per_step,
            len(aln_before[0]),
            len(aln_after[0]),
            outfile,
            plot_dir_prefix=plot_dir_prefix,
            file_prefix=file_prefix,
            included_steps=included_steps,
        )

        return self.return_filtered_output()

    def return_filtered_output(self):
        """
        Creates a temporary alignment file and a temporary population map file, writes data to them, and returns a new GenotypeData object with the filtered alignment.

        Returns:
            GenotypeData: A new GenotypeData object with the filtered alignment.

        Raises:
            None
        """
        # Create a temporary file and write some data to it
        aln = tempfile.NamedTemporaryFile(delete=False)

        if self.sample_indices is None:
            self.sample_indices = len(range(self.popgenio.samples))

        popmap = {k: v for k, v in self.popgenio.popmap.items() if k in self.samples}

        if self.popgenio.filetype == "vcf":
            vcf_attributes = self.popgenio.subset_vcf_data(
                self.loci_indices,
                self.sample_indices,
                self.popgenio.vcf_attributes,
                samples=self.samples,
                chunk_size=self.popgenio.chunk_size,
                is_filtered=True,
            )
        else:
            vcf_attributes = self.popgenio._vcf_attributes

        aln_filename = aln.name

        self.popgenio.write_phylip(
            aln_filename, snp_data=self.alignment, samples=self.samples
        )
        aln.close()

        popmap = tempfile.NamedTemporaryFile(delete=False)
        popmap_filename = popmap.name

        if self.sample_indices is not None:
            self.popgenio.popmap = {
                k: v for k, v in self.popgenio.popmap.items() if k in self.samples
            }

        with open(popmap_filename, "w") as fout:
            for key, value in self.popgenio.popmap.items():
                fout.write(f"{key}\t{value}\n")
        popmap.close()

        inputs = self.popgenio.inputs
        inputs["popmapfile"] = popmap_filename
        inputs["filename"] = aln_filename
        inputs["filetype"] = "phylip"
        inputs["verbose"] = False
        inputs["is_subset"] = True

        # Create a new object with the filtered alignment.
        new_popgenio = GenotypeData(**inputs)

        new_popgenio.filetype = self.popgenio.filetype
        new_popgenio.verbose = self.popgenio.verbose

        if self.popgenio.filetype == "vcf":
            new_popgenio.vcf_header = self.popgenio.vcf_header
            new_popgenio.vcf_attributes = vcf_attributes

        # When done, delete the file manually using os.unlink
        os.unlink(aln_filename)
        os.unlink(popmap_filename)

        return new_popgenio

    def calc_missing_proportions(
        self,
        alignment_array,
        missing_chars=["N", "-", ".", "?"],
        calculate_stdev=False,
        is_sample_filter=False,
    ):
        """Calculates the proportion of missing data in each column or row of the alignment.

        Args:
            alignment_array (numpy.ndarray): The alignment array.

            missing_chars (list, optional): List of characters representing missing data. Defaults to ["N", "-", ".", "?"].

            calculate_stdev (bool, optional): Whether to calculate the standard deviation of the missing data proportions. Defaults to False.

            is_sample_filter (bool, optional): Whether the calculation is for a sample filter. If True, calculates proportions per column (axis=1). If False, calculates proportions per row (axis=0). Defaults to False.

        Returns:
            numpy.ndarray or tuple: The proportion of missing data per column or row. If calculate_stdev is True, returns a tuple with the proportions and the standard deviation.

        Raises:
            None
        """
        axis = 1 if is_sample_filter else 0

        new_missing_counts = np.sum(np.isin(alignment_array, missing_chars), axis=axis)

        # Calculate the mean missing data proportion among all the columns
        missing_prop = new_missing_counts / alignment_array.shape[axis]

        if calculate_stdev:
            std_missing_prop = np.std(new_missing_counts / alignment_array.shape[axis])

        res = (missing_prop, std_missing_prop) if calculate_stdev else missing_prop

        return res

    def filter_missing(self, threshold, alignment=None, return_props=False):
        """Filters out columns with missing data proportion greater than the given threshold.

        Args:
            threshold (float): The maximum missing data proportion allowed.

            alignment (MultipleSeqAlignment, optional): The alignment to be filtered. Defaults to the stored alignment.

            return_props (bool, optional): Whether to return the mean missing data proportion among all columns after filtering. Defaults to False.

        Returns:
            MultipleSeqAlignment or tuple: The filtered alignment. If return_props is True, returns a tuple with the filtered alignment, the mean missing data proportion, and None.

        Raises:
            TypeError: If threshold is not a float value.

            ValueError: If threshold is not between 0.0 and 1.0 inclusive.
        """
        if alignment is None:
            alignment = self.alignment

        alignment_array = alignment

        missing_counts = np.sum(np.isin(alignment_array, ["N", "-", ".", "?"]), axis=0)
        mask = missing_counts / alignment_array.shape[0] <= threshold

        # Get the indices of the True values in the mask
        mask_indices = [i for i, val in enumerate(mask) if val]

        # Apply the mask to filter out columns with a missing proportion greater than the threshold
        filtered_alignment_array = alignment_array[:, mask]

        if return_props:
            missing_prop = self.calc_missing_proportions(filtered_alignment_array)
            return (
                filtered_alignment_array,
                missing_prop,
                None,
            )
        else:
            return filtered_alignment_array, mask_indices

    def filter_missing_pop(
        self, max_missing, alignment, populations=None, return_props=False
    ):
        """Filters out sequences from the alignment that have more than a given proportion of missing data in any given population.

        Args:
            max_missing (float): The maximum missing data proportion allowed.

            alignment (MultipleSeqAlignment): The alignment to be filtered.

            populations (dict, optional): A dictionary mapping population names to sample IDs. Defaults to None.

            return_props (bool, optional): Whether to return the mean and standard deviation of missing data proportions for each population after filtering. Defaults to False.

        Returns:
            MultipleSeqAlignment or tuple: The filtered alignment. If return_props is True, returns a tuple with the filtered alignment, a dictionary of mean missing data proportions for each population, and a dictionary of standard deviations of missing data proportions for each population.

        Raises:
            None
        """
        if populations is None:
            populations = self.popmap_inverse

        alignment_array = alignment

        def missing_data_proportion(column, indices):
            missing_count = sum(column[i] in {"N", "-", ".", "?"} for i in indices)
            return missing_count / len(indices)

        def exceeds_threshold(column):
            missing_props = {}
            flaglist = []
            for pop, sample_ids in populations.items():
                # Get sampleID indices for given population, if not removed
                # with filter_missing_sample
                indices = [i for i, sid in enumerate(sample_ids) if sid in self.samples]
                missing_prop = missing_data_proportion(column, indices)

                if missing_prop >= max_missing:
                    flagged = True
                else:
                    missing_props[pop] = missing_prop
                    flagged = False
                flaglist.append(flagged)
            return flaglist, missing_props

        mask_and_missing_props = np.array(
            [exceeds_threshold(col) for col in alignment_array.T], dtype=object
        )
        mask = np.all(
            np.array([mmp[0] for mmp in mask_and_missing_props], dtype=bool),
            axis=1,
        )

        # Get the indices of the True values in the mask
        mask_indices = [i for i, val in enumerate(mask) if not val]

        filtered_alignment_array = alignment_array[:, ~mask]

        if return_props:
            mean_missing_props = [mmp[1] for mmp in mask_and_missing_props]

            key_values = {
                key: [d.get(key) for d in mean_missing_props if d.get(key) is not None]
                for key in set().union(*mean_missing_props)
            }

            missing_props = {k: v for k, v in key_values.items()}
            std_missing_props = {k: np.std(v) for k, v in key_values.items()}
            return (
                filtered_alignment_array,
                missing_props,
                std_missing_props,
            )
        else:
            return filtered_alignment_array, mask_indices

    def filter_missing_sample(self, threshold, alignment=None, return_props=False):
        """Filters out sequences with missing data proportion greater than the given threshold.

        Args:
            threshold (float): The maximum missing data proportion allowed for each sequence.

            alignment (MultipleSeqAlignment, optional): The alignment to be filtered. Defaults to the stored alignment.

            return_props (bool, optional): Whether to return the mean missing data proportion among all sequences after filtering. Defaults to False.

        Returns:
            MultipleSeqAlignment or tuple: The filtered alignment. If return_props is True, returns a tuple with the filtered alignment, the mean missing data proportion among all sequences, and the indices of the filtered sequences.

        Raises:
            TypeError: If threshold is not a float value.

            ValueError: If threshold is not between 0.0 and 1.0 inclusive.
        """

        if alignment is None:
            alignment = self.alignment

        alignment_array = alignment

        missing_counts = np.sum(np.isin(alignment_array, ["N", "-", ".", "?"]), axis=1)

        mask = missing_counts / alignment_array.shape[1] <= threshold

        # Apply the mask to filter out sequences with a missing proportion greater than the threshold
        filtered_alignment_array = alignment_array[mask, :]

        # Get the indices of the True values in the mask
        mask_indices = [i for i, val in enumerate(mask) if val]

        # Convert the filtered alignment array back to a list of SeqRecord objects
        filtered_alignment = [
            filtered_alignment_array[i, :] for i, index in enumerate(mask_indices)
        ]

        if return_props:
            missing_prop = self.calc_missing_proportions(
                filtered_alignment_array, is_sample_filter=True
            )
            return filtered_alignment, missing_prop, mask_indices
        else:
            return filtered_alignment, mask_indices

    def filter_minor_allele_frequency(
        self, min_maf, alignment=None, return_props=False
    ):
        """Filters out loci (columns) where the minor allele frequency is below the threshold.

        Args:
            min_maf (float): The minimum minor allele frequency allowed.

            alignment (MultipleSeqAlignment, optional): The alignment to be filtered. Defaults to the stored alignment.

            return_props (bool, optional): Whether to return the mean missing data proportion among all columns after filtering. Defaults to False.

        Returns:
            MultipleSeqAlignment or tuple: The filtered alignment. If return_props is True, returns a tuple with the filtered alignment, the mean missing data proportion among all columns, and the minor allele frequencies.

        Raises:
            TypeError: If min_maf is not a float value.

            ValueError: If min_maf is not between 0.0 and 1.0 inclusive.
        """
        if alignment is None:
            alignment = self.alignment
        alignment_array = alignment

        def count_bases(column):
            base_count = {
                "A": 0,
                "C": 0,
                "G": 0,
                "T": 0,
            }
            for base in column:
                if base in base_count:
                    base_count[base] += 1
                elif base not in {"N", "-", ".", "?"}:
                    try:
                        ambig_bases = SeqUtils.IUPACData.ambiguous_dna_values[base]
                        for ambig_base in ambig_bases:
                            base_count[ambig_base] += 1
                    except KeyError:
                        pass
            return base_count

        def minor_allele_frequency(column):
            counts = count_bases(column)
            # Remove counts of "N", "-", and "." characters from the counts dictionary
            counts = {
                base: count
                for base, count in counts.items()
                if base not in {"N", "-", ".", "?"}
            }

            if not counts or all(v == 0 for v in counts.values()):
                return 0

            # Sort the counts by their values
            sorted_counts = sorted(counts.values(), reverse=True)
            total = sum(sorted_counts)

            # Calculate the frequencies of each allele
            freqs = [count / total for count in sorted_counts]

            # Return the frequency of the second most common allele (the minor allele)
            return freqs[1] if len(freqs) > 1 else 0

        maf = np.apply_along_axis(minor_allele_frequency, 0, alignment_array)
        mask = maf >= min_maf

        # Get the indices of the True values in the mask
        mask_indices = [i for i, val in enumerate(mask) if val]
        filtered_alignment_array = alignment_array[:, mask]

        if return_props:
            missing_prop = self.calc_missing_proportions(filtered_alignment_array)
            return (
                filtered_alignment_array,
                missing_prop,
                maf,
            )
        else:
            return filtered_alignment_array, mask_indices

    def filter_non_biallelic(self, threshold=None, alignment=None, return_props=False):
        """Filters out loci (columns) that are not biallelic.

        Args:
            threshold (None, optional): Not used.

            alignment (MultipleSeqAlignment, optional): The alignment to be filtered. Defaults to the stored alignment.

            return_props (bool, optional): Whether to return additional information. Defaults to False.

        Returns:
            MultipleSeqAlignment or tuple: The filtered alignment. If return_props is True, returns a tuple with the original missing data proportions, filtered missing data proportions, and the mask indicating the biallelic columns.

        Raises:
            None
        """

        if alignment is None:
            alignment = self.alignment

        # Convert the input alignment to a numpy array of sequences
        alignment_array = alignment

        iupac = {
            "R": ("A", "G"),
            "Y": ("C", "T"),
            "S": ("G", "C"),
            "W": ("A", "T"),
            "K": ("G", "T"),
            "M": ("A", "C"),
        }

        def count_unique_bases(column):
            """
            Args:
                column (str): A column of bases from an alignment.

            Returns:
                int: The number of unique bases in the column, excluding ambiguous and missing bases.
            """
            base_count = {
                "A": 0,
                "C": 0,
                "G": 0,
                "T": 0,
                "U": 0,
            }

            for base in column:
                if base in base_count:
                    base_count[base] += 1
                elif base in iupac:
                    base1, base2 = iupac[base]
                    base_count[base1] += 1
                    base_count[base2] += 1
                # Ignore "N", "-", and "." bases

            return len([count for count in base_count.values() if count > 0])

        unique_base_counts = np.apply_along_axis(count_unique_bases, 0, alignment_array)
        mask = unique_base_counts == 2

        # Get the indices of the True values in the mask
        mask_indices = [i for i, val in enumerate(mask) if val]

        # Apply the mask to filter non-biallelic columns
        filtered_alignment_array = alignment_array[:, mask]

        if return_props:
            orig_missing_prop = self.calc_missing_proportions(alignment_array)
            filt_missing_prop = self.calc_missing_proportions(filtered_alignment_array)
            return (
                orig_missing_prop,
                filt_missing_prop,
                mask,
            )
        else:
            return filtered_alignment_array, mask_indices

    def count_iupac_alleles(self, column):
        """Counts the number of occurrences of each IUPAC ambiguity code in a column of nucleotide sequences.

        Args:
            column (str): A string representing a column of nucleotide sequences.

        Returns:
            dict: A dictionary with the counts of the unambiguous nucleotide bases.

        Raises:
            None
        """
        iupac = {
            "A": "A",
            "C": "C",
            "G": "G",
            "T": "T",
            "U": "T",
            "R": "AG",
            "Y": "CT",
            "S": "GC",
            "W": "AT",
            "K": "GT",
            "M": "AC",
            "B": "CGT",
            "D": "AGT",
            "H": "ACT",
            "V": "ACG",
            "N": "ACGT",
        }

        counts = {"A": 0, "C": 0, "G": 0, "T": 0}

        for base in column:
            if base in iupac:
                for allele in iupac[base]:
                    counts[allele] += 1

        return counts

    def filter_monomorphic(self, threshold=None, alignment=None, return_props=False):
        """Filters out monomorphic sites from an alignment.

        Args:
            alignment (Bio.Align.MultipleSeqAlignment): The alignment to be filtered.

        Returns:
            filtered_alignment (Bio.Align.MultipleSeqAlignment): The filtered alignment.

        Raises:
            ValueError: If no loci remain in the alignment.
        """

        if alignment is None:
            alignment = self.alignment

        def is_monomorphic(column):
            """
            Determines if a column in an alignment is monomorphic.

            Args:
                column: a list of bases representing a column in an alignment

            Returns:
                A boolean indicating whether the column is monomorphic.
            """
            column_list = column.tolist()
            alleles = set(column_list)

            # Remove any ambiguity code
            alleles.discard("N")

            # Count the number of valid alleles
            valid_alleles = [
                allele for allele in alleles if allele not in ["-", ".", "?"]
            ]

            return len(valid_alleles) <= 1

        alignment_array = alignment.astype(str)

        if alignment_array.shape[1] > 0:
            mask = np.apply_along_axis(is_monomorphic, 0, alignment_array)
            filtered_alignment_array = alignment_array[:, ~mask]

            # Get the indices of the True values in the mask
            mask_indices = [i for i, val in enumerate(mask) if not val]

        else:
            raise ValueError(
                "No loci remain in the alignment. Try adjusting the filtering paramters."
            )

        if return_props:
            orig_missing_prop = self.calc_missing_proportions(alignment_array)
            filt_missing_prop = self.calc_missing_proportions(filtered_alignment_array)
            return (
                orig_missing_prop,
                filt_missing_prop,
                mask,
            )
        else:
            return filtered_alignment_array, mask_indices

    @staticmethod
    def resolve_ambiguity(base):
        """
        Resolves an IUPAC ambiguity code to the set of possible nucleotides it represents.

        Args:
            base (str): A single IUPAC character.

        Returns:
            set: A set of possible nucleotides represented by the IUPAC character.
        """
        iupac_dict = {
            "A": {"A"},
            "C": {"C"},
            "G": {"G"},
            "T": {"T"},
            "U": {"T"},
            "R": {"A", "G"},
            "Y": {"C", "T"},
            "S": {"G", "C"},
            "W": {"A", "T"},
            "K": {"G", "T"},
            "M": {"A", "C"},
            "B": {"C", "G", "T"},
            "D": {"A", "G", "T"},
            "H": {"A", "C", "T"},
            "V": {"A", "C", "G"},
            "N": {"A", "C", "G", "T"},
            "-": {"-"},
        }
        return iupac_dict.get(base.upper(), {"N"})

    def filter_singletons(self, threshold=None, alignment=None, return_props=False):
        """
        Filters out singletons from an alignment.

        Args:
            alignment (Bio.Align.MultipleSeqAlignment): The alignment to be filtered.

        Returns:
            filtered_alignment (Bio.Align.MultipleSeqAlignment): The filtered alignment.
        """

        if alignment is None:
            alignment = self.alignment

        def is_singleton(column):
            """
            Determines if a column in an alignment is a singleton.

            Args:
                column: a list of bases representing a column in an alignment

            Returns:
                A boolean indicating whether the column is a singleton, meaning that there is only one
            variant in the column and it appears only once.
            """
            column_list = column.tolist()
            alleles = {
                allele for allele in column_list if allele not in ["N", "-", ".", "?"]
            }
            allele_count = {allele: column_list.count(allele) for allele in alleles}

            if len(alleles) == 2:
                min_allele = min(alleles, key=lambda x: allele_count[x])
                return allele_count[min_allele] == 1
            return False

        alignment_array = alignment.astype(str)

        if alignment_array.shape[1] > 0:
            mask = np.apply_along_axis(is_singleton, 0, alignment_array)
            filtered_alignment_array = alignment_array[:, ~mask]
            # Get the indices of the True values in the mask
            mask_indices = [i for i, val in enumerate(mask) if not val]
        else:
            raise ValueError(
                "No loci remain in the alignment. Try adjusting the filtering paramters."
            )

        if return_props:
            orig_missing_prop = self.calc_missing_proportions(alignment_array)
            filt_missing_prop = self.calc_missing_proportions(filtered_alignment_array)
            return (
                orig_missing_prop,
                filt_missing_prop,
                mask,
            )
        else:
            return filtered_alignment_array, mask_indices

    def get_population_sequences(self, population):
        """
        Returns a list of sequence strings for a specific population.

        Args:
            population: str, the name of the population to retrieve sequences for.

        Returns:
            population_sequences: list, a list of sequence strings for the specified population.

        Raises:
            ValueError: If the specified population is not found in the object's list of populations.
        """
        population_indices = [
            i for i, pop in enumerate(self.populations) if pop == population
        ]
        alignment_array = self.alignment
        population_sequences = alignment_array[population_indices, :]
        return population_sequences.tolist()

    @staticmethod
    def print_filtering_report(
        before_alignment, after_alignment, loci_removed_per_step
    ):
        """
        Print a filtering report to the terminal.

        Args:
            before_alignment (list): The original alignment before filtering.

            after_alignment (list): The alignment after filtering.

            loci_removed_per_step (list of tuples): A list of tuples, where each tuple contains the name of a filtering step and the number of loci removed during that step.

        Returns:
            None.

        Raises:
            ValueError: If there is no data left after filtering, which could indicate an issue with the filtering or with the provided filtering parameters.

        Note:
            The function also raises a warning if none of the filtering arguments were changed from their defaults, in which case the alignment will not be filtered.
        """
        num_loci_before = len(before_alignment[0])
        num_samples_before = len(before_alignment)
        num_loci_after = len(after_alignment[0])
        num_samples_after = len(after_alignment)
        samples_removed = num_samples_before - num_samples_after

        def missing_data_percent(msa):
            total = len(msa) * len(msa[0])
            if total == 0:
                # for name, loci_removed in loci_removed_per_step:
                #     print(f"  {name}: {loci_removed}")
                raise ValueError(
                    "There is no data left after filtering. This can indicate an issue with the filtering or with the provided filtering parameters."
                )
            missing = np.count_nonzero(np.isin(msa, ["N", "-", ".", "?"]))
            return (missing / total) * 100

        missing_data_before = missing_data_percent(before_alignment)
        missing_data_after = missing_data_percent(after_alignment)

        print("Filtering Report:")
        print(f"  Loci before filtering: {num_loci_before}")
        print(f"  Samples before filtering: {num_samples_before}")

        if (
            all([x[1] == 0 for x in loci_removed_per_step])
            and samples_removed == 0
            and missing_data_before == missing_data_after
        ):
            warnings.warn(
                "The alignment was unchanged. Note that if none of the filtering arguments were changed from defaults, the alignment will not be filtered."
            )

        for name, loci_removed in loci_removed_per_step:
            print(f"  {name}: {loci_removed}")
        print(f"  Samples removed: {samples_removed}")
        print(f"  Loci remaining: {num_loci_after}")
        print(f"  Samples remaining: {num_samples_after}")
        print(f"  Missing data before filtering: {missing_data_before:.2f}%")
        print(f"  Missing data after filtering: {missing_data_after:.2f}%\n\n")

    def plot_missing_data_thresholds(
        self,
        output_file,
        num_thresholds=5,
        num_maf_thresholds=10,
        max_maf_threshold=0.2,
        show_plot_inline=False,
        plot_dir_prefix="snpio",
        file_prefix=None,
        plot_fontsize=28,
        plot_ticksize=20,
        plot_ymin=0.0,
        plot_ymax=1.0,
        plot_legend_loc="upper left",
        plot_format="png",
        dpi=300,
    ):
        """
        Plots the missing data and MAF proportions for different filtering thresholds.

        Args:
            output_file (str): The name of the output plot file.
            num_thresholds (int, optional): The number of thresholds to use for filtering. Defaults to 5.

            num_maf_thresholds (int, optional): The number of minor allele frequency (MAF) thresholds to use for filtering. Defaults to 10.
            max_maf_threshold (float, optional): The maximum MAF threshold to use for filtering. Defaults to 0.2.

            show_plot_inline (bool, optional): Whether to show the plot inline. Defaults to False.

            plot_dir_prefix (str, optional): The prefix of the directory to save the plot. Defaults to "snpio".

            file_prefix (str, optional): Prefix of the output filename. If ``file_prefix`` is None, then no prefix is prepended to the filename. Defaults to None.

            plot_fontsize (int, optional): The fontsize for plot labels. Defaults to 28.

            plot_ticksize (int, optional): The fontsize for plot ticks. Defaults to 20.

            plot_ymin (float, optional): The minimum y-axis value for the plot. Defaults to 0.0.

            plot_ymax (float, optional): The maximum y-axis value for the plot. Defaults to 1.0.

            plot_legend_loc (str, optional): The location of the plot legend. Defaults to "upper left".

            plot_format (str, optional): Format to save plot to. Supported image formats include: "pdf", "svg", "png", and "jpeg" (or "jpg"). Defaults to "png".

            dpi (int, optional): DPI resolution of plot. Defaults to 300.

        Returns:
            None.

        Raises:
            None.
        """
        thresholds = np.linspace(0.1, 1, num=num_thresholds, endpoint=True)
        maf_thresholds = np.linspace(
            0.0, max_maf_threshold, num=num_maf_thresholds, endpoint=True
        )
        sample_missing_data_proportions = []
        global_missing_data_proportions = []
        population_missing_data_proportions = []
        maf_per_threshold = []
        maf_props_per_threshold = []
        mask_indices = []

        for threshold, maf_threshold in zip(thresholds, maf_thresholds):
            mask_idx, sample_missing_prop = self.filter_per_threshold(
                self.filter_missing_sample,
                threshold,
                self.alignment,
                return_props=True,
                is_maf=True,
            )

            global_missing_prop = self.filter_per_threshold(
                self.filter_missing,
                threshold,
                self.alignment,
                return_props=True,
            )

            pop_missing_props = self.filter_per_threshold(
                self.filter_missing_pop,
                threshold,
                self.alignment,
                populations=self.popmap_inverse,
                return_props=True,
            )

            # Get MAF for each threshold
            maf_freqs, maf_props = self.filter_per_threshold(
                self.filter_minor_allele_frequency,
                maf_threshold,
                self.alignment,
                is_maf=True,
                return_props=True,
            )

            sample_missing_data_proportions.append(sample_missing_prop)
            global_missing_data_proportions.append(global_missing_prop)
            population_missing_data_proportions.append(pop_missing_props)
            maf_per_threshold.append(maf_freqs)
            maf_props_per_threshold.append(maf_props)

            # For sample-level filtering.
            mask_indices.append(mask_idx)

        (
            mono_orig_props,
            mono_filt_props,
            mono_mask,
        ) = self.filter_per_threshold(
            self.filter_monomorphic,
            self.alignment,
            is_bool=True,
            return_props=True,
        )

        (
            bi_orig_props,
            bi_filt_props,
            biallelic_mask,
        ) = self.filter_per_threshold(
            self.filter_non_biallelic,
            self.alignment,
            is_bool=True,
            return_props=True,
        )

        (
            sing_orig_props,
            sing_filt_props,
            singleton_mask,
        ) = self.filter_per_threshold(
            self.filter_singletons,
            self.alignment,
            is_bool=True,
            return_props=True,
        )

        def generate_df(props, thresholds, dftype, orig_props=None, mask=None):
            flattened_proportions = []
            flattened_thresholds = []

            if (orig_props is None and mask is not None) or (
                orig_props is not None and mask is None
            ):
                raise TypeError(
                    "orig_props and mask must both be either NoneType or not NoneType"
                )

            if orig_props is None:
                p = deepcopy(props)
            else:
                p = deepcopy(orig_props)

            if mask is None:
                for i, (threshold, array) in enumerate(zip(thresholds, p)):
                    flattened_proportions.extend(array)
                    flattened_thresholds.extend([f"{threshold:.2f}"] * len(array))

                df = pd.DataFrame(
                    {
                        "Threshold": flattened_thresholds,
                        "Proportion": flattened_proportions,
                        "Type": dftype,
                    }
                )

            else:
                df = pd.DataFrame(
                    {
                        "Proportion": orig_props,
                        "Filtered": mask,
                        "Type": dftype,
                    }
                )

            return df

        def generate_population_df(props, thresholds):
            df_list = []
            for threshold, prop_dict in zip(thresholds, props):
                for population, proportions in prop_dict.items():
                    temp_df = pd.DataFrame(
                        {
                            "Threshold": [f"{threshold:.2f}"] * len(proportions),
                            "Proportion": proportions,
                            "Type": [population] * len(proportions),
                        }
                    )
                    df_list.append(temp_df)

            df = pd.concat(df_list, ignore_index=True)
            return df

        df_sample = generate_df(sample_missing_data_proportions, thresholds, "Sample")

        df_global = generate_df(global_missing_data_proportions, thresholds, "Global")

        df_populations = generate_population_df(
            population_missing_data_proportions, thresholds
        )

        df_maf = generate_df(maf_props_per_threshold, maf_thresholds, "MAF")

        df_mono = generate_df(
            mono_filt_props,
            thresholds,
            "Monomorphic",
            mono_orig_props,
            mono_mask,
        )

        df_biallelic = generate_df(
            bi_filt_props,
            thresholds,
            "Biallelic",
            bi_orig_props,
            biallelic_mask,
        )

        df_singleton = generate_df(
            sing_filt_props,
            thresholds,
            "Singleton",
            sing_orig_props,
            singleton_mask,
        )

        # combine the two dataframes
        df = pd.concat([df_sample, df_global])
        df2 = pd.concat([df_mono, df_biallelic, df_singleton])

        Plotting.plot_filter_report(
            df,
            df2,
            df_populations,
            df_maf,
            maf_per_threshold,
            maf_props_per_threshold,
            output_file,
            plot_fontsize,
            plot_ticksize,
            plot_ymin,
            plot_ymax,
            plot_legend_loc,
            show_plot_inline,
            plot_dir_prefix=plot_dir_prefix,
            file_prefix=file_prefix,
            plot_format=plot_format,
            dpi=dpi,
        )

    def filter_per_threshold(
        self, filter_func, *args, is_maf=False, is_bool=False, **kwargs
    ):
        """
        Filters the alignment using the provided filter function for multiple thresholds.

        Args:
            filter_func (callable): The filtering function to apply.

            args: Positional arguments to pass to the filter function.

            is_maf (bool, optional): Indicates whether the filter is for minor allele frequency (MAF). Defaults to False.

            is_bool (bool, optional): Indicates whether the filter returns boolean values. Defaults to False.

            kwargs: Keyword arguments to pass to the filter function.

        Returns:
            tuple: A tuple containing the filtered results based on the filter function.
                If is_bool is True:
                    (orig_props, filt_props, mask)
                If is_maf is True:
                    (freqs, props)
                Otherwise:
                    props

        Raises:
            None.
        """
        if is_bool:
            orig_props, filt_props, mask = filter_func(*args, **kwargs)
            res = (orig_props, filt_props, mask)
        else:
            _, props, freqs = filter_func(*args, **kwargs)
            res = (freqs, props) if is_maf else props
        return res

    @property
    def alignment(self):
        """
        Gets the alignment data.

        Returns:
            numpy.ndarray: The alignment data as a numpy array.

        Raises:
            None.
        """
        if isinstance(self._alignment, MultipleSeqAlignment):
            a = np.array([list(str(record.seq)) for record in self._alignment])
        else:
            a = np.array(self._alignment)
        return a

    @alignment.setter
    def alignment(self, value):
        """
        Sets the alignment data.

        Args:
            value (MultipleSeqAlignment or numpy.ndarray): The alignment data to be set.

        Returns:
            None.

        Raises:
            None.
        """
        if isinstance(value, MultipleSeqAlignment):
            self._alignment = np.array([list(str(record.seq)) for record in value])
        else:
            self._alignment = value

    @property
    def msa(self):
        """
        Gets the multiple sequence alignment (MSA) data.

        Returns:
            The multiple sequence alignment (MSA) data.

        Raises:
            None.
        """
        return self._msa

    @msa.setter
    def msa(self, value):
        """
        Sets the multiple sequence alignment (MSA) data.

        Args:
            value: The multiple sequence alignment (MSA) data to be set.

        Returns:
            None.

        Raises:
            None.
        """
        self._msa = value

    @property
    def population_sequences(self):
        """
        Returns a dictionary of population sequences.

        The dictionary keys are the names of the populations, and the values are the corresponding sequences for each population.

        Sequences are in the form of a list of strings, where each string is a sequence for a given sample.

        Returns:
            dict: A dictionary of population sequences, where each key is the name of a population and the corresponding value is a list of sequences.

        Raises:
            None.
        """
        population_sequences = {}
        for population_name in self.populations:
            population_sequences[population_name] = self.get_population_sequences(
                population_name
            )
        return population_sequences

    @classmethod
    def print_cletus(cls):
        """Prints ASCII art of Cletus from the Simpsons (silly inside joke)."""
        # ASCII Cletus
        cletus_ascii = r"""
                                                      T                                            
                                                     M                                             
                                              +MI   :                                              
                                                  ?.M ?MM888DMD,                                   
                                                   M88888888888888DN.                              
                                                 M88888888888888888888N~                           
                                                8888888888888888888888888M                         
                                              ,88888888888888888888888888888O                      
                                             .88888888888888888888888888888888M                    
                                            =88888888888888888888888888888888888M                  
                                           M88888888888888888888888888888888888888M                
                                         D888888888N88888888888888888888888888888888O              
                                             M8888MID888888888888888888888888N8888888D             
                                            M8888MIIM8888M8888888888888888888MIIMD8888D.           
                                           7888MDIIIM888MII888888888888888888IIIIIN ~M8M           
                                           = IIIIIID88TTIIIM888DI8888M888888?IIIIIII.              
                                             MIIIMIIMIIIII?M88M ,888IIMDIDDIIIM      ?             
                                             M?IIIIIIIIII.      M8MMIIIIIIIII.        M            
                                             MINIIIIIIII       ,    NIIIIIII$     R    N           
                                          .MI?N8IIIIIIIN      &7    7IIIIIIM      M7  M           
                                         .IIIIIIIMIIIII=      &&     DIIIIIIIIM        D           
                                         MIIIIIIIIIIIIIM             ?IIIIIIIIIIM     M            
                                         ?IIIDIIIIIIIIIIN           MIIIIIIIIIIIIIN M.             
                                         IIIIIOIIIIIIIIIIM        ,IIIIIIIIIIIIIIIII$              
                                         7IIIIIMIIIIIIIIIII?NMMM8IMIIIIIIIIIIIIIIIIIIM             
                                         MIIIIIIIIIIIIIIIIMMMMM7IIIIIIIIIIIIIIIIIIIIIID            
                                         .IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIIM            
                                          NIIIIIIIIIIIIIIMIIIIIIIIIIIIIIIIIIIIMM?IIMM             
                                           NIIIIIIIIIIII7IIIIIIIIIMIIIIIIIIIIIIIIIIIIID.           
                                             IMMNIIIIIIIIIIIIIIII7IIIIMIIIMIIIIIII?II?IIIIIII8.    
                                                IIIIIIIIIIIIIIIIIIIIIIIIIIIIIIOIII?IIIIIIIIIII    
                                                IIIIIIIIINCLETUSD7IIIIIIIIIIIIIIIIIIIIIIIIIIIIM    
                                                IIIIIIMMMMMMM      O   IMM7IIIIIIIIIIIIIIIIID      
                                                IIIIMMMMMMMMMMM::MIM     N   .OMMZIIIIIDN,         
                                               :IIIMMMMMMMMMMMMIIIIM.,      M     ..             
                                               NIIINZDOZZMMMMMDIII~             MN~               
                                               IIIIMZZZZZZMMMIIIM                                 
                                              DIIIIIIMMMMZIIIIIIIIM                                
                                              IIIIIIIIIIIIIIIIIIIII,                               
                                             OIIIIIIIIIIIIIIIIIIIIN                                
                                             IIIIIIIIIIIIIIIIID..                                  
                                            MIIIIIIIIIIIIIIII,                                     
                                          ,IIIIIIIIIIIIIIIII?                                      
                                        DIIMIIIIIIIIIIIIIIIN                                       
                                    DOZZ7IIMIIIIIIIIIIIIIIO                                        
                                :M7MZZZLOLI7IIIIIIIIIIIIII~                                        
                             OYIIIIIZZZZMIIIIZIIIIIIIIIIIMN                                        
                         ,MIIIIIIIIIMZZZZMIIIIIMIIIIIIIIIIIM                                       
                       MIIIIIIIIIIIIIZZZZMMIIIIII7MMNMMIIIIZM                                      
                    MIIIIIIIIIIIIIIIINZZZOIIMIIIIIIIIIIIII7ZZIO                                    
                 .MIIIIIIIIIIIIIIIIIIMZZZZIIIINOIIIIIIIIIMZZZIIIM                                  
                   
        """
        print(cletus_ascii)



from setuptools import setup, find_packages

setup(
    name="snpio",
    version="1.0.3.3",
    url="https://github.com/btmartin721/SNPio",
    author="Bradley T. Martin and Tyler K. Chafin",
    author_email="evobio721@gmail.com",
    description="Reads and writes VCF, PHYLIP, and STRUCTURE files and performs data filtering on the alignment.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    license="GPL3",
    keywords=[
        "genomics",
        "bioinformatics",
        "population genetics",
        "SNP",
        "VCF",
        "PHYLIP",
        "STRUCTURE",
        "missing data",
        "filtering",
        "MAF",
        "biallelic",
    ],
    platforms=["Any"],
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.8",
    install_requires=[
        "biopython",
        "bokeh",
        "ete3",
        "holoviews",
        "kneed",
        "matplotlib",
        "numpy",
        "pandas",
        "panel",
        "plotly",
        "requests",
        "versioned-hdf5",
        "pysam",
        "scikit-learn",
        "scipy",
        "seaborn",
        "toytree",
        "kaleido",
        "psutil",
    ],
    extras_require={
        "docs": ["sphinx<7", "sphinx-rtd-theme", "sphinx-autodoc-typehints"],
        "intel": ["scikit-learn-intelex"],
        "dev": ["memory-profiler"],
    },
    entry_points={"console_scripts": ["snpio=run_snpio.py:main"]},
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    project_urls={
        "Source Code": "https://github.com/btmartin721/SNPio",
        "Bug Tracker": "https://github.com/btmartin721/SNPio/issues",
    },
)



import re

with open("all_packages.txt", "r") as file:
    lines = file.readlines()

# Regular expression pattern to match the package names
pattern = r"^\S+"

# Filter out the dependencies and keep only the top-level packages
top_level_packages = [
    re.match(pattern, line).group()
    for line in lines
    if not line.startswith("#")
]

# Write the top-level packages to a file
with open("top_level_packages.txt", "w") as file:
    file.write("\n".join(top_level_packages))
